#define OPTIONS "-d[GIMPST] -l[WE]"
/*
  Program: vw, a sentence generator for van Wijngaarden grammars.
  Author:  Dick Grune.
  Version: 821215
  Description:
	The program reads a VW grammar, followed by a list of starting symbols.
	All terminal productions are then generated. A production is terminal
	if no hyperrule applies.
	The only restriction is that no metanotion may be left-recursive.
	For details see vw.1
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<assert.h>
#include	<sys/types.h>
#include	<sys/times.h>
#define HZ	50	/* should be in sys/times.h */

/* 0. Definitions */
/* 0.1. Global declarations */

#define ERR		(-1)
#define NO_STRING	((char*)NULL)

#define YES		1
#define NO		0

static void CheckError(void);
static void ReportMetanotions(int, int, char *, char *);
static void ReportMetanotion(int, char *, char *);
static void Report(char *, char *);
static void PrintLetter(char);
static void PrintNumber(char *);

static void Clock(int);
static long TimeUsed(void);

static void Dump(char *, char *);
static void DumpNotion(char *, char *);

static char *NextRule(char *);
static char *NextAlternative(char *);

/* 0.2 Global definitions */
static char Note[] = "note";
static char Warning[] = "warning";
static char Error[] = "error";
static char Fatal[] = "fatal";

static char G_Flag = NO, I_Flag = NO, M_Flag = NO, P_Flag = NO,
	S_Flag = NO, T_Flag = NO;

static FILE *ifile;			/* input file */

/* 0.3. Memory management. */
/*	The whole grammar and all the sentential forms are kept in a single
	array starting at BeginMemory and ending at store_end; the first free
	is pointed to by StorePointer.
	The metarules begin at BeginMetarules.
	The hyperrules begin at BeginHyperrules and end at EndHyperrules.
*/

#define NO_SBRK		((char*)-1)	/* for sbrk failure */

#define StoreChar(ch)	if (StorePointer < store_end) *StorePointer ++= (ch); \
					else do_storec(ch)
static char *BeginMemory, *StorePointer, *store_end;
static char Inputbuf[BUFSIZ];		/* to prevent a call of malloc */

static void
InitStore(void) {
	/* Give Inputbuf to the file that is going to be used */
	setbuf(ifile, Inputbuf);
	setbuf(stdout, NO_STRING);	/* to prevent buffering of answers */
	BeginMemory = StorePointer = store_end = sbrk(0);
}

static void
do_storec(char ch) {
	char *nptr = sbrk(BUFSIZ);

	if (M_Flag) {
		printf("do_storec: BeginMemory = %lo, StorePointer = %lo, store_end = %lo, nptr = %lo\n",
			(long)BeginMemory, (long)StorePointer,
			(long)store_end, (long)nptr
		);
	}
	if (nptr == NO_SBRK) {
		Report("Out of core", Fatal);
	}
	assert(nptr == store_end);
	store_end = store_end + BUFSIZ;
	StoreChar(ch);
}

/* 1. The program. */

int
main(int argc, char *argv[]) {
	extern void Process(void);
	extern int options(int c, char *v[]);

	argc--;
	argv++;
	while (argc > 0 && argv[0][0] == '-' && argv[0][1] != '\0') {
		int nop = options(argc, argv);

		if (nop == ERR) {
			fprintf(stderr, "Options are [%s] [ file ]\n",
								OPTIONS);
			exit(ERR);
		}
		argc -= nop;
		argv += nop;
	}
	if (argc > 1) {
		fprintf(stderr, "Only one file allowed\n");
		exit(ERR);
	}
	if (argc == 0 || (argv[0][0] == '-' && argv[0][1] == '\0')) {
		ifile = stdin;
	}
	else {
		ifile = fopen(argv[0], "r");
		if (ifile == NULL) {
			fprintf(stderr, "Cannot open %s\n", argv[0]);
			exit(ERR);
		}
	}
	Process();
	return 0;
}

int
options(int c, char *v[]) {
	if (c < 1) return ERR;

	switch (v[0][1]) {
		char *ptr;

	case 'd':
		ptr = &v[0][2];
		while (*ptr) {
			switch (*ptr++) {
			case 'G':		/* displays Grammar */
				G_Flag = YES;
				break;
			case 'I':		/* display Input */
				I_Flag = YES;
				break;
			case 'M':		/* show Memory requests */
				M_Flag = YES;
				break;
			case 'P':		/* Print intermediate results */
				P_Flag = YES;
				break;
			case 'S':		/* display Substitutions */
				S_Flag = YES;
				break;
			case 'T':		/* Trace parser */
				T_Flag = YES;
				break;
			default:
				return ERR;
			}
		}
		return 1;

	case 'l':
		switch (v[0][2]) {
			extern char *ReportLevel;
		case 'W':
			ReportLevel = Warning;
			break;
		case 'E':
			ReportLevel = Error;
			break;
		default:
			return ERR;
		}
		return 1;

	default:
		return ERR;
	}
}

static int NumberOfProductions;

void
Process(void) {
	extern void ReadRules(void);
	extern int ReadStartwords(void);
	extern void Produce(void);

	InitStore();
	ReadRules();
	while(ReadStartwords()) {
		Clock(YES);
		while (NumberOfProductions > 0) {
			Produce();
		}
		Clock(NO);
		printf("\n");
	}
}
/* 2. The grammar. */

void
ReadRules(void) {
	extern void ReadMetarules(void);
	extern void ReadHyperrules(void);

	ReadMetarules();
	ReadHyperrules();
}

/* 2.1. Metarules. */

/*	The metarules are stored almost literally. A metaletter is identified
	by bit 8 being set. Bits 7 & 6 form a two-bit counter telling how
	many primes the letter carries and bits 5-0 are the rightmost 5
	bits of the letter.
*/

/* 2.1.1. Data structures. */
#define N_METAS			('Z'-'A'+1)

#define EOR(c) 	((c) == '.')				/* End Of Rule */
#define EOA(c) 	((c) == ';' || (c) == ':' || EOR(c))	/* End Of Alternative */
#define EON(c) 	((c) == ',' || EOA(c))			/* End Of Notion */

static char *m_rule[N_METAS];
#define Metarule(l)		m_rule[(l)&037]
#define MetaRightHandSide(l)	(Metarule(l) + 3 /* X::x */)
#define PackMetaletter(l,np)	(((l)-'A')|((np)<<5)|0200)
#define IsMetaletter(ch)	((ch)&0200)
#define Metaletter(ch) 		(((ch)&037)+'A')
#define PrimesCount(ch)		(((ch)>>5)&03)
#define ForAllMetanotions(c)	for (c = PackMetaletter('A',0); \
						Metaletter(c) <= 'Z'; c++)

static char *BeginMetarules = NO_STRING;

static char m_stat[N_METAS];
#define MetaStatus(l)		m_stat[(l)&037]

#define DEFINED			(1<<0)
#define USED			(1<<1)
#define FINITE		 	(1<<2)
#define EMPTY			(1<<3)
#define NON_LEFT_RECURSIVE	(1<<4)
#define CONTAMINATED		(1<<5)
#define IF_SOME_ALTERNATIVE	1
#define IF_EACH_ALTERNATIVE	2
/* In the present implementation NON_LEFT_RECURSIVE is set for those
   metanotions of which all alternatives start either with a terminal or
   with a metanotion for which NON_LEFT_RECURSIVE is set.
*/

/* 2.1.2. Read the metarules. */
void
ReadMetarules(void) {
	extern int GetMetarule(void);
	extern void CheckMetarules(void);

	BeginMetarules = StorePointer;
	while (GetMetarule()) {
		/* skip */
	}
	if (G_Flag) {
		Dump("of metarules", BeginMemory);
	}

	CheckMetarules();
}

void
CheckMetarules(void) {
	extern void TransitiveClosure(int, int, int (*)(char *));
	extern int IsFinite(char *);
	extern int IsEmpty(char *);
	extern int IsNonLeftrecursive(char *);
	extern int IsContaminated(char *);

	ReportMetanotions(DEFINED|USED, USED, "is used but not defined",
					Error);
	CheckError();

	TransitiveClosure(FINITE, IF_SOME_ALTERNATIVE, IsFinite);
	ReportMetanotions(DEFINED|FINITE, DEFINED,
					"has no finite production", Error);
	CheckError();

	TransitiveClosure(EMPTY, IF_SOME_ALTERNATIVE, IsEmpty);
	TransitiveClosure(NON_LEFT_RECURSIVE, IF_EACH_ALTERNATIVE,
					IsNonLeftrecursive);
	TransitiveClosure(CONTAMINATED, IF_SOME_ALTERNATIVE, IsContaminated);
	ReportMetanotions(DEFINED|CONTAMINATED, DEFINED|CONTAMINATED,
					"involves left-recursion", Note);
}

int
GetMetarule(void) {
	extern InputMark(void);

	char *metarule = StorePointer; /* before InputMark() */
	int letter = InputMark();
	int ch;

	if (EOR(letter)) return NO;
	if (!IsMetaletter(letter)) {
		ReportMetanotion(letter, "is no legal metaletter", Fatal);
	}
	if (MetaStatus(letter) & DEFINED) {
		ReportMetanotion(letter, "declared before", Error);
	}
	if (PrimesCount(letter) > 0) {
		Report("Prime in metarule definition", Error);
	}
	if (InputMark() != ':' || InputMark() != ':') {
		Report("Colons missing", Error);
	}
	while (ch = InputMark(), !EOR(ch)) {
		if (ch == ',') {
			Report("Comma in metarule", Error);
		}
		if (ch == ':') {
			Report("Colon in metarule", Error);
		}
		if (IsMetaletter(ch)) {
			if (PrimesCount(ch) > 0) {
				Report("Primes in metarule", Error);
			}
			MetaStatus(ch) |= USED;
		}
	}
	CheckError();
	Metarule(letter) = metarule;
	MetaStatus(letter) |= DEFINED;
	return YES;
}

/* 2.1.3. Check the metarules. */
void
TransitiveClosure(int bit, int type, int (*func)(char *)) {
	int done = NO;

	while (!done) {
		unsigned char ch;

		done = YES;
		ForAllMetanotions(ch) {
			if (	(MetaStatus(ch) & DEFINED)
			&&	!(MetaStatus(ch) & bit)
			) {
				char *meta_alt =
						MetaRightHandSide(ch);
				int some = NO, each = YES;

				do {
					if ((*func)(meta_alt)) {
						some = YES;
					}
					else	each = NO;
					meta_alt = NextAlternative(meta_alt);
				} while (meta_alt[-1] == ';');

				if (	(type == IF_SOME_ALTERNATIVE && some)
				||	(type == IF_EACH_ALTERNATIVE && each)
				) {
					MetaStatus(ch) |= bit;
					done = NO;
				}
			}
		}
	}
}

int
IsFinite(char *meta_alt) {
	char ch;

	while (ch = *meta_alt++, !EOA(ch)) {
		if (IsMetaletter(ch) && !(MetaStatus(ch) & FINITE)) return NO;
	}
	return YES;
}

int
IsEmpty(char *meta_alt) {
	char ch;

	while (ch = *meta_alt++, !EOA(ch)) {
		if (!IsMetaletter(ch) || !(MetaStatus(ch) & EMPTY)) return NO;
	}
	return YES;
}

int
IsNonLeftrecursive(char *meta_alt) {
	char ch;

	while (ch = *meta_alt++, !EOA(ch)) {
		if (!IsMetaletter(ch)) return YES;
		if (!(MetaStatus(ch) & NON_LEFT_RECURSIVE)) return NO;
		if (!(MetaStatus(ch) & EMPTY)) return YES;
	}
	return YES;
}

int
IsContaminated(char *meta_alt) {
	char ch;

	while (ch = *meta_alt++, !EOA(ch)) {
		if (	IsMetaletter(ch)
		&&	(	!(MetaStatus(ch) & NON_LEFT_RECURSIVE)
			||	(MetaStatus(ch) & CONTAMINATED)
			)
		)	return YES;
	}
	return NO;
}
/* 2.2. Hyperrules. */
/* 2.2.1. Data structures. */
static char *BeginHyperrules = NO_STRING;

/*	We shall need two lists of meatanotions, one while reading the
	grammar, and one while doing substitution.  Not too elegantly,
	these two have been combined in one array.
*/
#define MAXNOTIONS	(4*N_METAS)	/* every capital 4 times */
struct metanotion {
	char letter;
	char used;	/* during scanning */
	char *fi;	/* first in metaproduction, while substituting */
	char *fo;	/* first not in metaproduction, while substituting */
};

static struct metanotion  metanotion[MAXNOTIONS];
#define NO_NOTION		((struct metanotion *)0)
#define FirstMetanotion()	(&metanotion[0])
#define PushMetanotion()	(metapointer++)
#define PopMetanotion()		(--metapointer)
#define TopMetanotion()		(metapointer-1)
static struct metanotion *metapointer = FirstMetanotion();

static struct metanotion *
MetanotionWithLetter(char c) {
	struct metanotion *meta;

	for (meta = FirstMetanotion(); meta <= TopMetanotion(); meta++) {
		if (meta->letter == c) return meta;
	}
	return NO_NOTION;
}

/* 2.2.2. Reading the hyperrules. */
void
ReadHyperrules(void) {
	extern int GetHyperrule(void);

	BeginHyperrules = StorePointer;
	while (GetHyperrule()) {
		/* skip */
	}
	if (G_Flag) {
		Dump("of hyperrules", BeginHyperrules);
	}
}

int
GetHyperrule(void) {
	extern InputMark(void);
	extern void GetRightHandSide(void);

	int ch = InputMark();

	if (EOR(ch)) return NO;
	/* Read the LHS while keeping track of the metaletters */
	do {
		if (IsMetaletter(ch)) {
			struct metanotion *meta =
						MetanotionWithLetter(ch);

			MetaStatus(ch) |= USED;
			if (!(MetaStatus(ch) & DEFINED)) {
				ReportMetanotion(ch,
						"is not declared", Error);
			}
			if (MetaStatus(ch) & CONTAMINATED) {
				ReportMetanotion(ch,
					" in the LHS needs left-recursion, which is not implemented", Error);
			}

			if (meta == NO_NOTION) {
				/* new metanotion */
				meta = PushMetanotion();
				meta->letter = ch;
				meta->used = NO;
			}
			else {	/* old metanotion */
				meta->used = YES;
			}
		}
	} while (ch = InputMark(), !EON(ch));
	if (ch != ':') {
		Report("Missing colon", Error);
	}

	GetRightHandSide();
	return YES;
}

static char *EndHyperrules;
static char *FirstProduction = NO_STRING;
		/* Start of Production area. Active alternative is on top */
static char *Notion;
		/* Start of first non-terminal word; terminal words */
		/* are followed by a space (otherwise comma) */

int
ReadStartwords(void) {
	extern void GetRightHandSide(void);
	char *ptr;

	EndHyperrules = FirstProduction = StorePointer;
	GetRightHandSide();
	if (EOR(*FirstProduction)) return NO;

	NumberOfProductions = 1;
	for (ptr = FirstProduction; !EOR(*ptr); ptr++) {
		if (EOA(*ptr)) {
			*ptr = '.';
			NumberOfProductions++;
		}
	}
	if (G_Flag) {
		Dump("of startwords", FirstProduction);
	}
	ReportMetanotions(DEFINED|USED, DEFINED,
				"is defined but never used", Note);
	return YES;
}

void
GetRightHandSide(void) {
	extern InputMark(void);

	int ch;

	/* Read the RHS while testing the metaletters */
	while (ch = InputMark(), !EOR(ch)) {
		if (ch == ':') {
			Report("Illegal separator", Error);
		}
		if (IsMetaletter(ch)) {
			struct metanotion *meta =
						MetanotionWithLetter(ch);

			MetaStatus(ch) |= USED;
			if (!(MetaStatus(ch) & DEFINED)) {
				ReportMetanotion(ch, "is not declared", Error);
			}
			if (meta == NO_NOTION) {
				ReportMetanotion(ch, "occurs in the RHS only", Note);
			}
			else	meta->used = YES;
		}
	}
	CheckError();

	/* Remove metanotion list */
	while (TopMetanotion() >= FirstMetanotion()) {
		struct metanotion *meta = TopMetanotion();

		if (!meta->used) {
			ReportMetanotion(meta->letter,
				"never used in this hyperrule", Note);
		}
		PopMetanotion();
	}
}
/* 3. The production mechanism. */
void
Produce(void) {
	extern int ExpandFreeMetanotions(void);
	extern void DoHyperrules(void);
	extern void DeleteProduction(void);

	int previous = NumberOfProductions;
	char *term_ptr;

	/* Find the first non-terminal in the active sentential form */
	Notion = FirstProduction;
	for (term_ptr = FirstProduction; !EOR(*term_ptr); term_ptr++) {
		if (*term_ptr == ' ') {
			Notion = term_ptr+1;
		}
		else
		if (*term_ptr == ',') break;
	}
	/* Treat it */
	if (!ExpandFreeMetanotions()) {
		DoHyperrules();
	}
	/* Did the treatment add productions? */
	if (NumberOfProductions == previous) {
		if (EOR(*term_ptr)) {
			extern void PrintNotion(char *);

			Clock(NO);
			printf("%6ld sec: ", TimeUsed());
			PrintNotion(FirstProduction);
			Clock(YES);
			DeleteProduction();
		}
		else	*term_ptr = ' ';
	}
	else	DeleteProduction();
}

/* 3.1. Developping free metanotions. */
int
ExpandFreeMetanotions(void) {	/* develop free metanotions, if present */
	char *ptr, *term_ptr = NextRule(Notion) - 1;

	/* Scan the active notion for metaletters */
	for (ptr = Notion; !EON(*ptr); ptr++) {
		if (IsMetaletter(*ptr)) {
			char metaletter = *ptr;
			char *meta_alt = MetaRightHandSide(metaletter);

			if (T_Flag) {
				Clock(NO);
				printf("ExpandFreeMetanotions\n");
				DumpNotion("Notion", Notion);
				printf("\nmetaletter ");
				PrintLetter(metaletter);
				printf("\n");
				DumpNotion("meta_alt", meta_alt);
				printf("\n");
				Clock(YES);
			}

			/* Metanotion pad must be clean */
			PushMetanotion();
			assert(TopMetanotion() == FirstMetanotion());
			TopMetanotion()->letter = metaletter;

			/*	Substitute all meta_alternatives in the
				active notion
			*/
			do {
				extern void ProtectMetaletters(char *, char);
				extern void Substitute(char *, char *);
				extern void UnprotectMetaletter(char *);

				char *closeptr =
						NextAlternative(meta_alt) - 1;

				ProtectMetaletters(meta_alt, metaletter);
				TopMetanotion()->fi = meta_alt;
				TopMetanotion()->fo = closeptr;
				Substitute(Notion, term_ptr);
				UnprotectMetaletter(meta_alt);
				meta_alt = closeptr+1;
			} while (meta_alt[-1] == ';');
			PopMetanotion();
			if (P_Flag) {
				Clock(NO);
				Dump("from ExpandFreeMetanotions",
							FirstProduction);
				Clock(YES);
			}
			return YES;
		}
	}
	return NO;
}

void
ProtectMetaletters(char *meta_alt, char metaletter) {
	extern void ProtectMetaletter(char *, char *, char);
	char *ptr;

	for (ptr = meta_alt; !EOA(*ptr); ptr++) {
		if (IsMetaletter(*ptr)) {
			ProtectMetaletter(ptr, meta_alt, metaletter);
		}
	}
}

static char IndependencyMessageGiven = NO;

void
ProtectMetaletter(char *ptr, char *meta_alt, char metaletter) {
	int npr;

	/* Try an increasing number of primes */
	for (npr = 0; ; npr++) {
		extern int IsAvailableMetaletter(char, char *, char *, char);

		char newletter =
				PackMetaletter(Metaletter(*ptr), npr);

		if (	IsAvailableMetaletter(newletter,
				meta_alt, ptr, metaletter)
		) {
			*ptr = newletter;
			break;
		}
		if (npr == 3) {
			/* Even 3 were not enough */
			if (!IndependencyMessageGiven) {
				Clock(NO);
				printf("While substituting from ");
				PrintNumber(meta_alt);
				printf(", too many independent occurrences of ");
				PrintLetter(metaletter);
				printf(", independence sacrificed.\n");
				Clock(YES);
				IndependencyMessageGiven = YES;
			}
			break;
		}
	}
}

int
IsAvailableMetaletter(
	char newletter, char *meta_alt, char *term_ptr, char metaletter
) {
	char *ptr;

	/* If it has occurred already, it is out */
	for (ptr = meta_alt; ptr < term_ptr; ptr++) {
		if (*ptr == newletter) return NO;
	}
	/* If not, and it is the first occurrence, it is OK */
	if (newletter == metaletter) return YES;

	/* Otherwise, if it occurs anywhere, it is out */
	for (ptr = Notion; !EOA(*ptr); ptr++) {
		if (*ptr == newletter) return NO;
	}
	/* but if it doesn't, it is OK */
	return YES;
}

void
UnprotectMetaletter(char *meta_alt) {
	char *ptr;

	for (ptr = meta_alt; !EOA(*ptr); ptr++) {
		if (IsMetaletter(*ptr)) {
			*ptr = PackMetaletter(Metaletter(*ptr), 0);
		}
	}
}

/* 3.2. Apply hyperrules. */
void
DoHyperrules(void) {
	extern void DoHyperrule(char *, char *);

	char *hyp_ptr;

	/*	Try to parse the active notion according to all hyperrules
		successively
	*/
	for (	hyp_ptr = BeginHyperrules;
		hyp_ptr != EndHyperrules;
		hyp_ptr = NextRule(hyp_ptr)
	) {
		if (T_Flag) {
			printf("At 1\n");
		}
		DoHyperrule(hyp_ptr, Notion);
	}
	if (P_Flag) {
		Clock(NO);
		Dump("from DoHyperrules", FirstProduction);
		Clock(YES);
	}
}

/*	The following rules are concerned with the production of all parsings
	of the head of a string according to a given grammar.
*/
#define NO_TASK		((struct task *)NULL)
struct task {		/* 'pointer' to rest of proposed parse */
	struct task *next;
	union {
		struct metanotion *meta;	/* if next == NO_TASK */
		char *meta_alt;			/* if next != NO_TASK */
	} first;
};

void
DoHyperrule(char *hyp_ptr, char *prod_ptr) {
	/*	A persistent attempt is made to parse the string starting
		at prod_ptr according to the pattern starting at hyp_ptr
	*/
	char letter = *hyp_ptr;

	if (T_Flag) {
		Clock(NO);
		printf("DoHyperrule\n");
		DumpNotion("hyp_ptr", hyp_ptr);
		DumpNotion("prod_ptr", prod_ptr);
		printf("\n");
		Clock(YES);
	}
	if (IsMetaletter(letter)) {
		extern void DoMetaletter(char *, char *, char, struct task *);

		struct metanotion *meta =
					MetanotionWithLetter(letter);

		if (meta == NO_NOTION) {
			/*	New metanotion; all productions of 'letter'
				must be found that match zero or more
				characters of the string starting at prod_ptr
			*/
			struct task task;

			meta = PushMetanotion();
			meta->letter = letter;
			meta->fi = prod_ptr;
			/* meta->fo will set when a suitable parse is found */
			task.first.meta = meta;
			task.next = NO_TASK;
			if (T_Flag) {
				printf("At 2\n");
			}
			DoMetaletter(hyp_ptr, prod_ptr, letter, &task);
			PopMetanotion();
		}
		else {
			/*	Old metanotion, for which a value is already
				proposed; check the value
			*/
			char *fi = meta->fi, *fo = meta->fo;

			while (fi < fo && *fi == *prod_ptr) {
				fi++;
				prod_ptr++;
			}
			if (fi == fo) {
				if (T_Flag) {
					printf("At 3\n");
				}
				DoHyperrule(hyp_ptr+1, prod_ptr);
			}
			else { /* parse failed */ }
		}
	}
	else
	if (*hyp_ptr == ':') {
		/* Hyperrule exhausted, and so should the production be */
		if (EON(*prod_ptr)) {
			/* Caught a parse! */
			extern void Substitute(char *, char *);

			Substitute(hyp_ptr+1, prod_ptr);
		}
		else { /* parse failed */ }
	}
	else
	if (*hyp_ptr == *prod_ptr) {
		/* The easy case */
		if (T_Flag) {
			printf("At 4\n");
		}
		DoHyperrule(hyp_ptr+1, prod_ptr+1);
	}
	else { /* parse failed */ }
}

static void
DoMetaAlternative(
	char *hyp_ptr, char *prod_ptr, char *meta_alt, struct task *task
) {
	char letter = *meta_alt;

	if (T_Flag) {
		Clock(NO);
		printf("DoMetaAlternative\n");
		DumpNotion("hyp_ptr", hyp_ptr);
		DumpNotion("prod_ptr", prod_ptr);
		DumpNotion("meta_alt", meta_alt);
		printf("\n");
		Clock(YES);
	}
	if (EOA(letter)) {
		/* Meta_alternative finished */
		if (task->next == NO_TASK) {
			/* we have found a production for the metanotion */
			/* note its extent */
			task->first.meta->fo = prod_ptr;
			if (T_Flag) {
				printf("At 6\n");
			}
			/* and continue to parse the rest */
			DoHyperrule(hyp_ptr+1 /* skip over metaletter */,
								prod_ptr);
		}
		else {
			/*	continue the attempted production of the
				metaletter in *hyp_ptr
			*/
			if (T_Flag) {
				printf("At 7\n");
			}
			DoMetaAlternative(hyp_ptr, prod_ptr,
					task->first.meta_alt, task->next);
		}
	}
	else
	if (IsMetaletter(letter)) {
		/*	The meta_alternative contains a metaletter for which
			again a suitable matching production must be found.
			Parsing the rest of this alternative will be our
			next task
		*/
		extern void DoMetaletter(char *, char *, char, struct task *);

		struct task newtask;

		newtask.first.meta_alt = meta_alt+1;
		newtask.next = task;
		if (T_Flag) {
			printf("At 8\n");
		}
		DoMetaletter(hyp_ptr, prod_ptr, letter, &newtask);
	}
	else
	if (*prod_ptr == letter) {
		/* The easy case */
		if (T_Flag) {
			printf("At 9\n");
		}
		DoMetaAlternative(hyp_ptr, prod_ptr+1, meta_alt+1, task);
	}
	else { /* parse failed */ }
}

void
DoMetaletter(char *hyp_ptr, char *prod_ptr, char letter, struct task *task) {
	/* Parsing is attempted for all alternatives of the metanotion */
	char *meta_alt = MetaRightHandSide(letter);

	if (T_Flag) {
		Clock(NO);
		printf("DoMetaletter\n");
		DumpNotion("hyp_ptr", hyp_ptr);
		DumpNotion("prod_ptr", prod_ptr);
		printf("letter: ");
		PrintLetter(letter);
		printf("\n");
		DumpNotion("meta_alt", meta_alt);
		printf("\n");
		Clock(YES);
	}

	do {
		char *closeptr = NextAlternative(meta_alt) - 1;
		if (T_Flag) {
			printf("At 5\n");
		}
		DoMetaAlternative(hyp_ptr, prod_ptr, meta_alt, task);
		meta_alt = closeptr + 1;
	} while (meta_alt[-1] == ';');
}

/* 3.3. Substitute hyperrule. */
void
Substitute(char *hyp_ptr, char *term_ptr) {
	/*	The current production is replaced by a series of copies,
		each with the active notion (as demarcated by Notion and
		term_ptr) replaced by another alternative of the hyperrule.
		Metanotions are replaced as specified in the metanotion list.
	*/
	int ch;

	if (S_Flag) {
		extern void PrintMetanotions(void);

		Clock(NO);
		printf("Substitution from ");
		PrintNumber(hyp_ptr);
		printf(":\n");
		PrintMetanotions();
		Clock(YES);
	}

	do {
		char *tp = term_ptr;
		char *ptr, ch1;

		/* Copy the dull part */
		for (ptr = FirstProduction; ptr < Notion; ptr++) {
			StoreChar(*ptr);
		}

		/*	Replace the active notion by the next alternative of
			the hyperrule
		*/
		while (ch = *hyp_ptr++, !EOA(ch)) {
			if (IsMetaletter(ch)) {
				struct metanotion *meta =
						MetanotionWithLetter(ch);

				if (meta == NO_NOTION) {
					StoreChar(ch);
				}
				else {
					char *p;

					for (p = meta->fi; p < meta->fo; p++) {
						StoreChar(*p);
					}
				}
			}
			else {
				StoreChar(ch);
			}
		}

		/* Copy the other notion in the active production */
		do	StoreChar(*tp);
		while (ch1 = *tp++, !EOR(ch1));

		NumberOfProductions++;
	} while (!EOR(ch));
	if (S_Flag) {
		Dump("from Substitute", FirstProduction);
	}
}
/* 4. Printing of results. */

void
PrintMetanotions(void) {
	/* Print the contents of the metanotion list */
	struct metanotion *meta;

	for (	meta = FirstMetanotion();
		meta <= TopMetanotion();
		meta++
	) {
		char *p;

		printf("  Metaletter ");
		PrintLetter(meta->letter);
		printf(" = <");
		for (p = meta->fi; p < meta->fo; p++) {
			PrintLetter(*p);
		}
		printf(">\n");
	}
}

void
PrintNotion(char *ptr) {
	do	PrintLetter(*ptr);
	while (!EOR(*ptr++));
}

static void
PrintLetter(char ch) {

	if (IsMetaletter(ch)) {
		int np = PrimesCount(ch);

		printf("%c", Metaletter(ch));
		while (np--) {
			printf("%c", '\'');
		}
	}
	else
		printf("%c", ch);
	if (EOR(ch)) {
		printf("%c", '\n');
	}
}

static void
PrintNumber(char *nptr) {
/* number of rule into which nptr points */
	int number = 0;
	char *ptr, *where;

	if (nptr >= StorePointer) {
		printf("high pointer nptr = %lo", (long)nptr);
		return;
	}
	else
	if (FirstProduction != NO_STRING && nptr >= FirstProduction) {
		where = "production";
		ptr = FirstProduction;
	}
	else
	if (BeginHyperrules != NO_STRING && nptr >= BeginHyperrules) {
		where = "hyperrule";
		ptr = BeginHyperrules;
	}
	else
	if (BeginMetarules != NO_STRING && nptr >= BeginMetarules) {
		where = "metarule";
		ptr = BeginMetarules;
	}
	else {
		printf("low pointer nptr = %lo", (long)nptr);
		return;
	}

	while (ptr <= nptr) {
		number++;
		while (!EOR(*ptr++)) {
			if (ptr == StorePointer) break;
		}
	}
	printf("%s #%d", where, number);
}

void
DeleteProduction(void) {
	char *p1 = FirstProduction, *p2 = NextRule(FirstProduction);

	assert(NumberOfProductions > 0);
	while (p2 < StorePointer) {
		*p1++ = *p2++;
	}
	StorePointer = p1;
	NumberOfProductions--;
}

static char *
NextRule(char *ptr) {
	while (!EOR(*ptr)) {
		ptr++;
	}
	return ptr + 1;
}

static char *
NextAlternative(char *ptr) {
	while (!EOA(*ptr)) {
		ptr++;
	}
	return ptr + 1;
}
/* 5. Input. */

static int LineCount = 1;

static int
InputChar(void) {
	int ch = getc(ifile);

	if (I_Flag) {
		printf("InputChar: ch = %c (= %3o)\n", ch&0177, ch);
	}
	if (ch == EOF) {
		Report("Premature EOF", Fatal);
	}
	if (ch == '\n') {
		LineCount++;
	}
	return ch;
}

static char NonCommentAhead = '\0';

static int
InputNonCommentChar(void) {
	int ch;

	if (NonCommentAhead > '\0') {
		ch = NonCommentAhead;
		NonCommentAhead = '\0';
	}
	else {
		while (ch = InputChar(),
			ch <= ' ' || ch >= 0177 || ch == '\\') {
			if (ch == '\\') {
				while (ch = InputChar(), ch != '\n') {
					/* skip */
				}
			}
		}
	}
	return ch;
}

void
BackNonCommentChar(char ch) {
	NonCommentAhead = ch;
}

int
InputMark(void) {
	int ch = InputNonCommentChar();

	if ('A' <= ch && ch <= 'Z') {
		extern void BackNonCommentChar(char);

		int np = 0, pc;

		while (pc = InputNonCommentChar(), pc == '\'') {
			np++;
		}
		BackNonCommentChar(pc);
		if (np > 3) {
			Report("More than 3 primes", Error);
		}
		ch = PackMetaletter(ch, np);
	}
	if (ch == '\'') {
		Report("Prime not appended to metanotion", Error);
	}
	StoreChar(ch);
	return ch;
}
/* 6. Error reporting. */
static char *LastLevel = NO_STRING;
char *ReportLevel = Note;

static void
ReportMetanotions(int mask, int bits, char *message, char *level) {
	unsigned char ch;

	if (level < ReportLevel) return;
	ForAllMetanotions(ch) {
		if ((MetaStatus(ch) & mask) == bits) {
			ReportMetanotion(ch, message, level);
		}
	}
}

static void
ReportMetanotion(int ch, char *message, char *level) {

	if (level < ReportLevel) return;
	PrintLetter(ch);
	printf("%c", ' ');
	Report(message, level);
}

static void
Report(char *message, char *level) {
	if (level < ReportLevel) return;
	printf("%s (%s), detected at line %d\n", message, level, LineCount);
	if (level > LastLevel) {
		LastLevel = level;
	}
	if (level == Fatal) {
		exit(1);
	}
}

static void
CheckError(void) {
	if (LastLevel == Error) {
		exit(1);
	}
}
/* 7. Miscellaneous routines. */
static time_t UtimeUsed = (time_t)0;
static char ClockStat = NO;

static void
Clock(int st) {
	struct tms tbuf;

	assert(ClockStat != st);
	times(&tbuf);
	UtimeUsed = tbuf.tms_utime - UtimeUsed;
	ClockStat = st;
}

static long
TimeUsed(void) {
	assert(ClockStat == NO);
	return UtimeUsed/HZ;
}

static void
Dump(char *idf, char *ptr) {
	printf("### Dump %s\n", idf);
	while (ptr < StorePointer) {
		PrintLetter(*ptr++);
	}
	printf("\n");
}

static void
DumpNotion(char *idf, char *ptr) {
	printf("%s: ", idf);
	PrintNotion(ptr);
}
